package memegene;

import java.util.Random;
import java.util.Vector;
/** Clase abstracta que contiene métodos generales para la implementación de un algoritmo genético.
 *
 * @author Mauricio Alarcón M.
 * @author Felipe Urra A.
 */

public abstract class GA{
    /** Contenedor vector con individuos de la población.*/
    public  Vector<Individuo> poblacion,npoblacion;
    /** Arreglo con la capacidad de los individuos*/
    public  float[]           capacidad;
    /** Cantidad de generaciones que durará el algoritmo*/
    public  int               generacion;
    /** Cantidad de genes en las cadenas de los individuos*/
    protected int             numGenes;
    /** Índice de posición que indica el mejor individuo*/
    protected int             elite;
    /** Cota menor y mayor del rango de valores que toma la variable de la función a evaluar*/
    protected float           a,b;
    /** Probabilidad en que los genes pueden mutar*/
    protected float           mutacion;
    /** Objeto de la clase Ruleta*/
    protected Ruleta          ruleta;
    /** Flags que indican si se está usando el método de selección elite y si se está minimizando o maximizando la función*/
    protected boolean         es_elite,min;

    /** ID de método de selección "Elite".*/
    final static int ELITE = 0;
    /** ID de método de selección "Ranking".*/
    final static int RANKING = 2;
    /** ID de método de selección "Torneo".*/
    final static int TOURNAMENT = 1;

    /** Constructor
     *
     * @param individuos Cantidad de individuos por generación.
     * @param pmut Probabilidad de mutar de sso genes de los individuos.
     * @param min  Flag que indica se está minimizando o maximizando.
     */
    GA(int individuos,float pmut,boolean min){

        this.poblacion = new Vector<Individuo>(individuos,1);
       
        this.capacidad  = new float[individuos];

        this.generacion = 1;
        this.elite      = 0;
        this.mutacion   = pmut;
        this.es_elite   = false;
        this.min        = min;

    }

    /** Setea el número de genes de los individuos
     * 
     * @param numGen número de genes
     */
    public void setNumGenes(int numGen){
        this.numGenes = numGen;
        int largo = capacidad.length;
        
        for(int i=0; i < largo; i++)
            poblacion.addElement(new Individuo(numGen));
    }

    //setea fitness de cada individuo de la poblacion y guarda en variable 'elite' el índice del
    //individuo con mayor capacidad
    /** Método que calcula la aptitud de los individuos.*/
    abstract public void fitness();

    /** Setea intervalo de la variable de la función a optimizar.
     *
     * @param a Cota menor del rango
     * @param b Cota mayor del rango
     */

    protected void setInterval(float a,float b){

        this.a = a;
        this.b = b;

    }
    /** Cruza inviduos de la población
     *
     * @param pos    Lugar en el que se cruzarán
     * @param index  Ubicación de individuos contiguos en el vector de población
     */
    protected void cruzamiento(int pos, int index){  //pos: lugar desde donde se cruzaran genes, index: individuos
        int gen1,gen2;

        for(int i=pos; i < this.numGenes; i++){
            gen1 = npoblacion.elementAt(index).genes[i];
            gen2 = npoblacion.elementAt(index+1).genes[i];

            if(gen1 != gen2){
                npoblacion.elementAt(index).genes[i] = gen2;
                npoblacion.elementAt(index+1).genes[i] = gen1;
            }
        }
    }

    /**
     * Método de selección Elite.
     */
    public void elite(){
        ruleta = new Ruleta(capacidad,true);
        es_elite = true;
        int[] ganadores;
        int posicion;

        npoblacion = new Vector<Individuo>(poblacion.size(),1);
        ganadores = ruleta.go();
        
        for(int i = 0; i < ganadores.length ; i += 2){
            posicion = ruleta.lugar(this.numGenes-1)+1;    //lugar de cruce, numero entre 1 y (numero de genes-1)

            npoblacion.addElement(poblacion.elementAt(ganadores[i]).clone());
            npoblacion.addElement(poblacion.elementAt(ganadores[i+1]).clone());

            cruzamiento(posicion,i);
        }

        this.generacion++;

    }
/**
 * Método que se encarga de la mutación.
 */
    public void mutacion(){
        int l;
        for(int i=0; i < npoblacion.size(); i++){
            for(l=0; l < this.numGenes ;l++)
                if(ruleta.go(mutacion)){

                    if(npoblacion.elementAt(i).genes[l] == 0)
                        npoblacion.elementAt(i).genes[l] = 1;
                    else
                        npoblacion.elementAt(i).genes[l] = 0;
            }
        }
        if(es_elite)
            insert();

        poblacion = npoblacion;
    }
/**
 * Inserta
 */
   protected void insert(){
   
        int l = ruleta.lugar(poblacion.size());
        Individuo ind = poblacion.elementAt(elite).clone();
        npoblacion.removeElementAt(l);
        npoblacion.insertElementAt(ind,l);

    }
    /** Método de selección por torneo.
     * 
     * @param probabilidad Probabilidad de victoria del individuo con mejor fitness
     */
    public void tournament(float probabilidad){
        int m,n,posicion;
        int iteraciones = poblacion.size();
        npoblacion = new Vector<Individuo>(iteraciones,1);
        ruleta = new Ruleta(capacidad,true);
        for(int i=0; i < iteraciones ;i++){
            do{
              m = ruleta.lugar(iteraciones);            //participante aleatorio
              n = ruleta.lugar(iteraciones);
            }while(m == n);
            //System.out.println("Concursantes: "+(m+1)+" "+(n+1));

            if(capacidad[m] > capacidad[n]){
                if(ruleta.go(probabilidad)){
                    npoblacion.addElement(poblacion.elementAt(m).clone());
                    //System.out.println("MEJOR");
                }
                else{
                    npoblacion.addElement(poblacion.elementAt(n).clone());
                    //System.out.println("PEOR");
                }
            }
            else
                if(ruleta.go(probabilidad)){
                    npoblacion.addElement(poblacion.elementAt(n).clone());
                    //System.out.println("MEJOR");
                }
                else{
                    npoblacion.addElement(poblacion.elementAt(m).clone());
                    //System.out.println("PEOR");
                }
            if(i%2==1){
                posicion = ruleta.lugar(this.numGenes-1)+1;   //posicion de cruce para genes
                cruzamiento(posicion,i-1);

            }
        }
        this.generacion++;
    }

    /**
     * Método de selección Ranking
     */
    public void ranking(){
        int ganadores[],posicion;
        npoblacion = new Vector<Individuo>(poblacion.size(),1);
        Sorter s = new Sorter(poblacion,capacidad);
        float ev[] = new float[capacidad.length];
        float max = 1.1f, min = 0.9f;
        s.sort();                                   //ordena poblacion de menor a mayor usando quicksort

        for(int i=0; i < poblacion.size() ; i++){    //posicion en el arreglo es su ranking
            ev[i] =(float) max - (max - min)*i/(poblacion.size()-1);
        }

        ruleta = new Ruleta(ev,false);
        ganadores = ruleta.go();

        for(int i = 0; i < ganadores.length ; i += 2){
            posicion = ruleta.lugar(this.numGenes-1)+1;    //lugar de cruce, numero entre 1 y (numero de genes-1)

            npoblacion.addElement(poblacion.elementAt(ganadores[i]).clone());
            npoblacion.addElement(poblacion.elementAt(ganadores[i+1]).clone());
                
         this.generacion++;
        }
    }

    /**
     * En caso de minimización, esta función ajusta los valores de la función para convertirlo en maximización.
     * @param in valor de entrada.
     * @return Valor escalado.
     */
    protected float ajustar(float in){
        if(in == 0)
            return 1f;
        else
            return 1/in;
    }

    /** Desparametriza el valor ajustado.
     *
     * @param in Valor a restaurar.
     * @return   Valor original.
     */
    protected float desajustar(float in){
      return 1/in - 1;

    }

    /**
     * Ejecuta el algoritmo genético.
     * @param iteraciones  Cantidad de generaciones que durará el algoritmo.
     * @param idSeleccion  ID para los differentes,
     * @param probabilidadTorneo  Probabilidad de victoria del individuo más apto en caso de método de selección por torneo
     *
     */
    public void ejecutar(int iteraciones,int idSeleccion, float probabilidadTorneo){
        
        for(int i=0; i < iteraciones;i++){
            this.fitness();

            switch(idSeleccion){
                case 0:
                       this.elite();
                       break;
                case 1:
                       this.tournament(probabilidadTorneo);
                       break;
                case 2:
                       this.ranking();

            }
            
            this.mutacion();
        }
        this.fitness();
    }


    /**
     * Función a optimizar.
     * @param in String que representa la función
     * @return Valor de la función
     */
    abstract float   funcion(float in);               

}
