/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package memegene;

import java.util.Random;
import org.nfunk.jep.JEP;

/**
 *
 * @author Mauricio Alarcón M.
 * @author Felipe Urra A.
 */
public class MAFunction extends MA{

    /**String que contiene la funcion a optimizar*/
    String funcion;
    /**Objeto JEP que se preocupa de parsear las funciones*/
    JEP evaluador  = new JEP();
    /**Parametro utilizado en la busqueda local, que indica cuantos bits se modificarán
    * para asi probar localmente mejores soluciones que la actual
     */
    int bitsACambiar;
    /**GPanel para ingresar datos al gráfico*/
    GPanel panel;
    /**Valores que se ingresan al gráfico*/
    double[] valores;

    /**
     * Constructor
     * @param individuos Numero de soluciones
     * @param pmut probabilidad de mutación
     * @param min boolean que es verdadero si se intenta minimizar la funcion
     * @param a rango inferior de la variable independiente
     * @param b rango superior de la variable independiente
     * @param precision numero de digitos suceres a la coma flotante de la variable independiente
     * @param panel GPanel del marco de la aplicacion
     */
    MAFunction(int individuos,float pmut,boolean min,float a,float b, int precision,GPanel panel){
        super(individuos,pmut,min);
        this.setNumGenes(this.calculaBits(a, b, precision));
        this.panel = panel;
        setInterval(a,b);
        evaluador.addStandardFunctions();
        evaluador.addStandardConstants();
        evaluador.addComplex();
        evaluador.addFunction("sen",new org.nfunk.jep.function.Sine());
        evaluador.addVariable("x",0);
        evaluador.setImplicitMul(true);  //permite 2x vez de 2*x        
        this.bitsACambiar = this.numGenes/2;
    }

    /**
     * Metodo que realiza, para cada individuo, una busqueda local para encontrar una mejor
     * solución.
     */
     public void busquedaLocal(){

         Random r = new Random();
         Individuo ind,indTemp;
         int i,size,j;
         int indiceCambiar;

         size = this.poblacion.size();
         for(i=0;i<size;i++){

            ind = this.poblacion.elementAt(i);
            indTemp = ind.clone();

            for(j=0;j<this.bitsACambiar;j++){

                indiceCambiar = r.nextInt(this.numGenes);  //el numero del bit a cambiar
                if(ind.genes[indiceCambiar] == 1){
                    indTemp.genes[indiceCambiar] = 0;
                }else{
                    indTemp.genes[indiceCambiar] = 1;
                }
            }
            if(this.min){
                if((funcion(indTemp.getFit())) < funcion(ind.getFit())){
                ind = indTemp;
            }
            }else{
                if((funcion(indTemp.getFit())) > funcion(ind.getFit())){
                ind = indTemp;
            }
            }
            
         }
     }


     /**Calcula el numero de bits para representar la variable independiente dentro de los paramentros de
      * entrada
      *
      * @param a rango inferior de la variable independiente
      * @param b rango superior de la variable independiente
      * @param precision numero de digitos sucesores a la coma flotante de la variable independiente
      * @return numero de bits del vector utilizado por el algoritmo memético
      */
    public int calculaBits(float a, float b, int precision){
        int bits;
        double factor = (b-a)*Math.pow(10,precision);
        for(bits=1; bits > -1; bits++)
            if(Math.pow(2,bits-1) < factor && factor <= (Math.pow(2,bits)-1))
                break;
            return bits;
    }

    //setea fitness de cada individuo de la poblacion y guarda en variable 'elite' el índice del
    //individuo con mayor capacidad
    /**
     * Setea el fitness de cada individuo y guarda en la variable "elite" el indice del individuo
     * de mejor capacidad.
     */
    public void fitness(){

        float mejor=0;
        elite = 0;
        int largo = capacidad.length;
        for(int i=0;i < largo;i++){
            poblacion.elementAt(i).setFit(a,b);
            if(min)                              //se ajusta capacidad en caso de minimización
              capacidad[i] = super.ajustar(funcion(poblacion.elementAt(i).getFit())); //cambiar por abstracta
            else
              capacidad[i] = funcion(poblacion.elementAt(i).getFit());          //cambiar por abstracta

            if(i==0) mejor = capacidad[i];    //elite
            else
              if(capacidad[i] > mejor){
                 //System.out.println("Iteracion: "+i+" min: "+min+" elite: "+elite+" capacidad: "+capacidad[i]);
                mejor = capacidad[i];
                elite = i;
              }
        }

    }
    /**
     * Setea la funcion a utilizar
     * @param funcion funcion a utilizar en formato String
     */
    void setFuncion(String funcion){
        evaluador.parseExpression(funcion);
    }

    /**
     * Evalua la funcion objetivo
     * @param in valor real de un individuo
     * @return variable dependiente de la funcion
     */
    float funcion(float in){               //pa que no tire pifia
        //aca tenemos un objeto JEP...y al setear
        evaluador.addVariable("x",in);
        return (float)evaluador.getValue();
    }    

  

    /**
     * Ejecuta el algoritmo memético
     * @param iteraciones numero de iteraciones
     * @param idSeleccion tipo de seleccion
     * @param probabilidadTorneo probabilidad de elegir el mejor individuo para el tipo de seleccion
     * tournament
     */
    public void ejecutar(int iteraciones,int idSeleccion, float probabilidadTorneo){

        System.out.println("EJECUTAR MA");
        System.out.println(iteraciones);
        this.valores = new double[iteraciones];
        for(int i=0; i < iteraciones;i++){        
            this.fitness();

            switch(idSeleccion){
                case 0:
                       this.elite();
                       break;
                case 1:
                       this.ranking();
                       break;
                case 2:
                       this.tournament(probabilidadTorneo);

            }
            this.mutacion();
            this.busquedaLocal();
            //System.out.println("MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM");
            
            
            
            valores[i] = funcion(poblacion.elementAt(elite).getFit());
            
            //go gráfico

        }
        this.fitness();
        this.panel.setSerie2(valores);
    }

        public void print(){

        System.out.println("-----Algoritmo Memético----\n");
        
        System.out.println("Mejor Individuo: "+poblacion.elementAt(elite).getFit());
        if(min){
          System.out.println("Salida: " +desajustar(capacidad[elite]));
        }
        else
          System.out.println("Salida: " +capacidad[elite]+"\n");
    }
    

}
