package memegene;

import java.util.Vector;
/** Clase que ordena los individuos y su fitness utilizando el algorito de ordenamiento <i>quicksort</i>.
 * @author Mauricio Alarcón M.
 * @author Felipe Urra A.
 */
public class Sorter {
  /** Arreglo con fitness de la población.*/
  public float[] arreglo;
  /** Contenedor con individuos.*/
  Vector<Individuo> vector;
  
  /** Constructor.
   *
   * @param poblacion Vector de individuos a ordenar.
   * @param capacidad Fitness de los individuos.
   */
  public Sorter(Vector<Individuo> poblacion, float[] capacidad) {
    this.arreglo = capacidad;
    this.vector = poblacion;
  }

  /** Método de ordenamiento <i>quicksort</i>.*/
  public void sort() {
    quicksort(0, vector.size()-1);
  }

  /** Ordenamiento <i>quicksort</i> recursivo.
   *
   * @param i  Índice izquierdo.
   * @param j  Índice derecho.
   */
  void quicksort(int i, int j) {
    if (i>=j)
      return;
    int k= particion(i, j);
    quicksort(i, k-1);
    quicksort(k+1, j);
  }
  /** Parte del algoritmo quicksort.
   * 
   * @param i Índice izquierdo.
   * @param j Índice derecho.
   * @return
   */
  int particion(int i, int j) { 
    int k= i;
    i++;
    for(;;) {
      while (i<j && compara(i,k)<=0)
        i++;
      while (i<j && compara(j,k)>=0)
        j--;
      if (i==j)
        break;
      swap(i,j);
    }
    if (compara(k,i)<0)
      i--;
    swap(k,i);
    return i;
  }

  /** Intercambia elemento i del arreglo con elemento j.
   *
   * @param i Índice de elemento a cambiar por elemento j.
   * @param j Índice de elemento a cambiar por elemento i.
   */
  public void swap(int i, int j) {
    float x= arreglo[i];
    arreglo[i]= arreglo[j];
    arreglo[j]= x;
    Individuo y = vector.elementAt(i).clone();
    Individuo z = vector.elementAt(j).clone();
    this.vector.removeElementAt(i);
    this.vector.insertElementAt(z,i);
    this.vector.removeElementAt(j);
    this.vector.insertElementAt(y,j);
  }
  /** Método que retorna -1 si elemento i del arreglo es menor que elemento j, 0 si son iguales y 1 en caso contrario.
   *
   * @param i
   * @param j
   * @return
   */
  public int compara(int i, int j) {
    if (arreglo[i] > arreglo[j]) return -1;
    else if (arreglo[i] == arreglo[j]) return 0;
    else return 1;
  }
  
}
